/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ReconnectingSocket.h"

#include <cassert>

#include "util.h"

typedef boost::system::error_code error_code;
typedef boost::asio::ip::tcp::resolver resolver;

namespace ds {
namespace vdsm {

ReconnectingSocket::ReconnectingSocket(boost::asio::io_service& ioService)
        : m_socket(ioService)
        , m_isActive(false)
        , m_resolver(ioService)
        , m_query("")
        , m_state(State::disconnected)
        , m_timer(ioService)
        , m_resolveTimeout(boost::chrono::seconds(20))
        , m_connectTimeout(boost::chrono::seconds(20))
        , m_idleTimeout(boost::chrono::seconds(20))
{
    VDSM_IENTER();
}

void ReconnectingSocket::setQuery(const resolver::query& query)
{
    VDSM_IENTER("%s:%s", query.host_name().c_str(), query.service_name().c_str());
    assert(!m_isActive);
    m_query = query;
    m_resolvedQueryIt = resolver::iterator();
}

void ReconnectingSocket::start()
{
    VDSM_IENTER();
    assert(!m_isActive);
    m_isActive = true;
    asyncResolve();
}

void ReconnectingSocket::reset()
{
    //public method
    VDSM_IENTER();
    resetImpl();
}

void ReconnectingSocket::resetImpl()
{
    VDSM_ENTER();
    assert(m_isActive);
    close();
    asyncWaitTimer(m_idleTimeout * (80 + randInt() % 40) / 100);
}

void ReconnectingSocket::close()
{
    VDSM_ENTER("m_state:%d", int(m_state));
    if (m_state == State::disconnected) {
        return;
    }
    auto oldState = m_state;
    m_state = State::disconnected;
    {
        error_code ec; //ignore close errors
        m_socket.shutdown(boost::asio::ip::tcp::socket::shutdown_both, ec);
        m_socket.close(ec);
    }
    m_timer.cancel();
    if (oldState == State::connected) {
        m_disconnected(); //notify higher code
    }
}

void ReconnectingSocket::asyncWaitTimer(const boost::chrono::milliseconds &timeout)
{
    VDSM_ENTER("timeout:%d", timeout.count());
    assert(m_isActive);

    m_timer.expires_from_now(timeout);
    m_timer.async_wait([this](const error_code& e)
    {
        if (e) {
            return; //timer was aborted
        }
        VDSM_ENTER();

        close();
        //Try next endpoint or resolve endpoints again
        if (m_resolvedQueryIt != resolver::iterator()) {
            asyncConnect();
        } else {
            asyncResolve();
        }
    });
}

void ReconnectingSocket::asyncResolve()
{
    VDSM_ENTER();
    m_resolver.async_resolve(m_query, [=](const error_code& e, resolver::iterator it)
    {
        if (e == boost::asio::error::operation_aborted) {
            return; //'this' may already be deleted
        }
        VDSM_ENTER("e:%s", e.message().c_str());
        if (e) {
            vdsm_report(LOG_WARNING, "Failed to resolve %s\n", m_query.host_name().c_str());
            resetImpl();
            return;
        }
        m_resolvedQueryIt = it;
        asyncConnect();
    });

    asyncWaitTimer(m_resolveTimeout);
}

void ReconnectingSocket::asyncConnect()
{
    VDSM_ENTER();

    if (m_resolvedQueryIt == resolver::iterator()) {
        resetImpl();
        return;
    } 

    m_state = State::connecting;
    auto endpoint = m_resolvedQueryIt->endpoint();
    m_resolvedQueryIt++;
    m_socket.async_connect(endpoint, [=](const error_code& e) 
    {
        if (e == boost::asio::error::operation_aborted) {
            return; //'this' may already be deleted
        }
        VDSM_ENTER("e:%s", e.message().c_str());
        if (e) {
            vdsm_report(LOG_WARNING, "Failed to connect to %s:%d e:%s\n",
                endpoint.address().to_string().c_str(), endpoint.port(),
                e.message().c_str()
            );
            //Preceed to next endpoint immediately.
            //Some endpoints can be unaccessible
            //and it should not trigger idle timeout
            m_socket.close();
            asyncConnect();
            return;
        }
        m_timer.cancel();
        m_state = State::connected;

        vdsm_report(LOG_NOTICE, "Connected to vdc at %s:%d\n",
            endpoint.address().to_string().c_str(), endpoint.port()
        );
        m_connected(); //notify higher code
    });

    asyncWaitTimer(m_connectTimeout);
}

}} //namespace ds::vdsm
