/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>
             Sergey Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.


    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __VDSM_VDC_API_H__
#define __VDSM_VDC_API_H__

#include "messages/messages.pb-c.h"
#include "datamodel.h"
#include "role.h"
#include "AsioVdcClient.h"
#include <pthread.h>


#define VDC_USER_PROTOBUF_MESSAGE_LENGTH    4096

typedef struct vdc_client {
    struct vdc_client *next;

    char dsuid[DSUID_STR_LEN];
    char *uri;
    char *port; /* only used in getaddrinfo, therefore a string */
    char *avahi_service_name;
    bool autodiscovered;

    void *asio; //Pointer to asio C++ object
} vdc_client_t;

/** list entry for a vdc containing a list of affected dsuids. */
typedef struct _vdc_dsuid_list_t {
    struct _vdc_dsuid_list_t *next;
    vdc_client_t *vdc_client;   ///< the vdc client
    char **dsuid;               ///< list of dsuids, ready for use in protobuf field
    size_t n_dsuid;             ///< size of dsuids list
} vdc_dsuid_list_t;

typedef enum {
    REQUEST_CONTEXT_NONE,
    REQUEST_CONTEXT_GET_PROPERTY_FID,
    REQUEST_CONTEXT_GET_PROPERTY_MODE,
    REQUEST_CONTEXT_GET_PROPERTY_LTMODE,
    REQUEST_CONTEXT_GET_PROPERTY_BUTTON_ID_GRP_0,
    REQUEST_CONTEXT_GET_PROPERTY_PBCHANNEL,
    REQUEST_CONTEXT_GET_PROPERTY_PBGROUP,
    REQUEST_CONTEXT_GET_PROPERTY_OUTPUT_VALUE,
    REQUEST_CONTEXT_GET_PROPERTY_SCENES,
    REQUEST_CONTEXT_GET_PROPERTY_NAME,
    REQUEST_CONTEXT_GET_PROPERTY_SENSORS,
    REQUEST_CONTEXT_GET_PROPERTY_BINARY_INPUTS,
    REQUEST_CONTEXT_GET_PROPERTY_SENSOR_VALUE,
    REQUEST_CONTEXT_GET_PROPERTY_GROUP_MEMBERSHIP,
    REQUEST_CONTEXT_GET_PROPERTY_OPC_VALUES,
    REQUEST_CONTEXT_GET_PROPERTY_OPC_SCENE_CONFIG,
    REQUEST_CONTEXT_GET_PROPERTY_OPC_SCENES_VALUE,
    REQUEST_CONTEXT_GET_PROPERTY_ZONEID,
    REQUEST_CONTEXT_GET_VDC_NAME,
    REQUEST_CONTEXT_GET_VDC_ZONDEID,
    REQUEST_CONTEXT_GET_VDC_CAPABILITIES,
} request_context_type_t;

typedef struct {
    request_context_type_t context;
    uint8_t offset;
    uint8_t count;
    uint8_t channel;
    uint8_t scene;
    uint8_t operation;
    uint16_t result;
} request_cache_t;

/**
 * vDc user protobuf message response event structure
 */
typedef struct {
    QEvent super;
    uint16_t size;
    uint8_t response[VDC_USER_PROTOBUF_MESSAGE_LENGTH];
} VdcUserProtobufMessageResponseEvt;

enum {
    VDC_API_OK = 0,

    VDC_API_ERR_OUT_OF_MEMORY = -1,
    VDC_API_ERR_SYSTEM = -2,
    VDC_API_ERR_PARAM = -3,
    VDC_API_ERR_NOT_CONNECTED = -4,
    VDC_API_ERR_SOCKET = -5,
    VDC_API_ERR_CLIENT_NOT_FOUND = -6
};

int vdc_api_init(const dsuid_t *dsmid, VDSMRole role, uint16_t listen_port);
int vdc_api_add_connection(const char* connection_url);
int vdc_api_process_discovered_vdc_host(const char* host, const char *port,
                                        const dsuid_t* dsuid,
                                        const char *service_name, bool noauto,
                                        bool collection_enabled);
void vdc_api_remove_auto_connection(const char *service_name);
bool vdc_api_in_client_list(const dsuid_t* dsuid);

void vdc_api_iterate();
void vdc_api_shutdown();

void vdc_api_free_message(Vdcapi__Message *msg);

int vdc_api_send_message_by_device_dsuid(const char *dev_dsuid,
                                         Vdcapi__Message *msg,
                                         request_cache_t *rc);

int vdc_api_send_message_by_vdc_dsuid(const char *vdc_dsuid,
                                      Vdcapi__Message *msg,
                                      request_cache_t *rc);

void vdc_api_send_get_property_by_device_dsuid(const char *dsuid,
                                               const char *name,
                                               request_cache_t *rc);

void vdc_api_send_get_property_by_vdc_dsuid(const char *vdc_dsuid,
                                            const char *name,
                                            request_cache_t *rc);

void vdc_api_send_get_properties_indexed_by_device_dsuid(const char *dsuid,
                                                         const char *name,
                                                         uint32_t* array,
                                                         uint32_t array_len,
                                                         request_cache_t *rc);

int vdc_api_next_message_id();

int vdc_api_send_message(vdc_client_t *vdc_client, Vdcapi__Message *msg,
                          request_cache_t *rc);

void vdc_api_send_get_property(vdc_client_t *vdc_client, const char *dsuid,
                               const char *name,
                               request_cache_t *rc);

void vdc_api_send_get_properties_indexed(vdc_client_t *vdc_client, const char *dsuid,
                                         const char *name,
                                         uint32_t* array, uint32_t array_len,
                                         request_cache_t *rc);

void vdc_api_device_get_name(uint16_t device_id);
void vdc_api_device_set_name(uint16_t device_id, const char *name);


void vdc_api_device_set_opc_scene(uint16_t device_id, uint8_t OPCscene,
                                  uint8_t channel, uint16_t value);

bool vdc_api_device_set_opc_scene_config(uint16_t device_id, uint8_t OPCscene,
                                         uint16_t value);

bool vdc_api_device_set_opc_scene_dc_flags(uint16_t device_id, uint8_t OPCscene,
                                           uint16_t value);

void vdc_api_device_get_opc_scene_value(uint16_t device_id, uint8_t OPCscene,
                                        uint8_t channel);

void vdc_api_device_get_opc_scene_config(uint16_t device_id, uint8_t OPCscene,
                                         uint8_t OPCoperation);

void vdc_api_device_set_opc_output_state(uint16_t device_id, uint8_t channel,
                                         uint16_t value, bool apply);

void vdc_api_device_get_opc_output_state(uint16_t device_id, uint8_t channel);

void vdc_api_device_register_sensors(uint16_t device_id);
void vdc_api_device_register_opc_channels(uint16_t device_id);
void vdc_api_device_register_binary_inputs(uint16_t device_id);
void vdc_api_device_get_group_membership(uint16_t device_id);
void vdc_api_device_ping(uint16_t device_id);
void vdc_api_device_remove(uint16_t device_id);
bool vdc_api_device_set_register(uint16_t device_id, uint8_t param_bank,
                                 uint8_t paramOffset, uint8_t val);
void vdc_api_device_get_sensor_value(uint16_t device_id, uint8_t sensor_index);
void vdc_api_device_get_zone(const char* dsuid);

void vdc_api_zone_push_sensor_value(uint8_t local_zone_id, uint8_t group_id,
                                    uint8_t sensor_type, uint16_t sensor_value);

int vdc_api_get_vdc_dsuid(dsuid_t *devices, int length);
bool vdc_api_get_vdc_host_for_vdc(const char *dev_dsuid, vdc_client_t **vdc_client);

bool vdc_api_vdsm_holds_vdc(const dsuid_t *vdc_dsuid);

void vdc_api_device_send_announce_ok(uint32_t messageId, uint16_t deviceId);

/*---- vdc specific commands -----------------------------------------------*/
void vdc_api_vdc_send_announce_ok(uint32_t messageId, const char* vdc_dsuid);
void vdc_api_vdc_get_name(const char *vdcdsuid);
void vdc_api_vdc_set_name(const dsuid_t *vdcdsuid, const char *name);
void vdc_api_vdc_get_defaultZoneID(const char *vdc_str);
void vdc_api_vdc_set_defaultZoneID(const char *vdc_str, uint16_t zoneID);
void vdc_api_vdc_get_capabilities(const char *vdc_str);

/**
 * helper function: convert dsuid to device_dsuid_t
 * @dsuid [in] single device
 * @return device_dsuid_t (list with one entry)
 */
device_dsuid_t *singleDsuidList(char *dsuid);

/**
 * @brief resortByVdcClient transform a given input list of device dsuid
 *        to a vdc_dsuid_list_t. (device dsuid sorted by vdc clients)
 *
 * @param dsuid_list [in] a list of device dsuids
 * @param vdc_dsuid_list [out] a vdc list with an according dsuid list
 */
void resortByVdcClient(device_dsuid_t *dsuid_list, vdc_dsuid_list_t **vdc_dsuid_list);

/**
 * @brief free_vdc_item Free the memory for vdc_item
 * @param vdc_item [in] item to be deleted
 */
void free_vdc_item(vdc_dsuid_list_t *vdc_item);

void vdc_api_set_vanish_vdc(vdc_client_t *vdc_client);
void vdc_api_send_ping(vdc_client_t *vdc_client);
void vdc_api_send_hello(vdc_client_t *vdc_client);
void vdc_api_process_message(vdc_client_t *vdc_client,
    const unsigned char *data, const uint16_t len
);


#endif//__VDSM_VDC_API_H__
